<?php

if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

/**
 * Backlinks plugin CLI commands
 *
 * ## EXAMPLES
 *
 *     wp backlinks update
 *
 * @when after_wp_load
 */

class BacklinksCLI {

    private function display_posts( $header, $posts ) {
        $output = [];
		foreach ( $posts as $post ) {
			$output[] = [
				__('ID', 'backlinks-taxonomy') => $post->ID,
				__('Type', 'backlinks-taxonomy') => $post->post_type,
				__('Status', 'backlinks-taxonomy') => $post->post_status,
				__('Title', 'backlinks-taxonomy') => $post->post_title,
				__('Date', 'backlinks-taxonomy') => $post->post_date,
				// __('Link', 'backlinks-taxonomy') => get_permalink( $post ),
			];
		}

        if ( $output ) {
            WP_CLI::line( $header );
            WP_CLI\Utils\format_items( 'table', $output, join(',', array_keys($output[0])) );
        } else {
            WP_CLI::warning( sprintf( __( '%s - NONE', 'backlinks-taxonomy' ), $header ) );
        }
    }

    /**
     * Show configuration
     *
     * ## EXAMPLES
     *
     *     wp backlink config
     *
     * @alias thumb
     */

    public function config( $args, $assoc ) {
        global $backlinks_plugin;

        WP_CLI::line( "Post types: " . join(',', $backlinks_plugin->post_types()));
        WP_CLI::line( "Post status: " .join(',', $backlinks_plugin->post_statuses()));
    }



    /**
     * Update links for all posts
     *
     * ## OPTIONS
     *
     * [<post_id>...]
     * : Update only the given posts
     *
     * [--format=<format>]
     * : Render output in a particular format.
     * ---
     * default: table
     * options:
     *   - table
     *   - csv
     *   - json
     *   - count
     *   - yaml
     * ---
     *
     *
     * ## EXAMPLES
     *
     *     wp backlinks rescan
     *
     */

    public function rescan( $args, $assoc ) {
        global $backlinks_plugin;

        if ( empty( $args ) ) {
            $posts = get_posts( $backlinks_plugin->build_post_query() );
        } else {
            $posts = get_posts( $backlinks_plugin->build_post_query( [
                'post__in' => array_map( 'intval', array_filter( $args, 'intval' ) ),
                'post_status' => array_merge( [ 'draft' ], $backlinks_plugin->post_statuses() ),
            ] ) );
        }

		$output = [];

		foreach ( $posts as $post ) {
			$links = $backlinks_plugin->register_outlinks( $post );

			$output[] = [
				__('ID', 'backlinks-taxonomy') => $post->ID,
				__('Type', 'backlinks-taxonomy') => $post->post_type,
				__('Status', 'backlinks-taxonomy') => $post->post_status,
				__('Title', 'backlinks-taxonomy') => $post->post_title,
				__('Outgoing links', 'backlinks-taxonomy') => count( $links ),
			];
		}

		if ( $output ) {
			WP_CLI\Utils\format_items( $assoc['format'], $output, join(',', array_keys($output[0])) );
		}
    }



    /**
     * Show backlinks for all posts
     *
     * ## OPTIONS
     *
     * [--max=<limit>]
     * : Show only posts with less than <limit> backlinks
     *
     * [--format=<format>]
     * : Render output in a particular format.
     * ---
     * default: table
     * options:
     *   - table
     *   - csv
     *   - json
     *   - count
     *   - yaml
     * ---
     *
     *
     * ## EXAMPLES
     *
     *     wp backlinks status
     *
     */

    public function status( $args, $assoc ) {
        global $backlinks_plugin;

        $max = 9999;
        if ( isset( $assoc['max'] ) )
            $max = intval( $assoc['max'] );

        $posts = get_posts( $backlinks_plugin->build_post_query() );

		$pools = [];

		foreach ( $posts as $post ) {
			$outgoing = get_the_terms( $post, $backlinks_plugin->taxonomy() );
			if ( is_wp_error( $outgoing ) ) {
				WP_CLI::error( $outgoing );
				$outgoing = [];
			} elseif ( $outgoing === FALSE ) {
				$outgoing = [];
			}

            $incoming = $backlinks_plugin->get_backlinks( $post );

            $count = count( $incoming );
			$pools[ $count ][] = [
				__('ID', 'backlinks-taxonomy') => $post->ID,
				__('Type', 'backlinks-taxonomy') => $post->post_type,
				__('Status', 'backlinks-taxonomy') => $post->post_status,
				__('Title', 'backlinks-taxonomy') => $post->post_title,
				__('Incoming', 'backlinks-taxonomy') => $count,
				__('Outgoing', 'backlinks-taxonomy') => count( $outgoing ),
			];
        }

        ksort( $pools, SORT_NUMERIC );

        foreach ( $pools as $count => $output ) {
            if ( $count > $max )
                break;

            WP_CLI::line( sprintf( "Found %d posts with %d backlinks", count( $output ), $count ) );
            WP_CLI\Utils\format_items( $assoc['format'], $output, join(',', array_keys($output[0])) );
        }
    }


    /**
     * Show registed links for post
     *
     * ## OPTIONS
     *
     * <post_id>
     * : Show incoming and outgoing links for post_id
     *
     * [--format=<format>]
     * : Render output in a particular format.
     * ---
     * default: table
     * options:
     *   - table
     *   - csv
     *   - json
     *   - count
     *   - yaml
     * ---
     *
     * ## EXAMPLES
     *
     *     wp backlink show 42
     *
     * @alias thumb
     */

    public function show( $args, $assoc ) {
        global $backlinks_plugin;

        $posts = get_posts( $backlinks_plugin->build_post_query( [
            'post__in' => [ intval($args[0] ) ],
        ] ) );

		if ( !$posts ) {
			WP_CLI::error( 'No such eligible post' );
			return;
		}
        $post = $posts[0];

        $this->display_posts( __( 'This post', 'backlinks-taxonomy' ) , [ $post ] );

		$outgoing = $backlinks_plugin->get_outlinks( $post );
        $header = sprintf( __('Found %d outgoing links.', 'backlinks-taxonomy'), count( $outgoing ));
        $this->display_posts( $header , $outgoing );

		$incoming = $backlinks_plugin->get_backlinks( $post );
        $header = sprintf( __('Found %d incoming links.', 'backlinks-taxonomy'), count( $incoming ));
        $this->display_posts( $header , $incoming );
    }



    /**
     * Suggest links for post
     *
     * ## OPTIONS
     *
     * <post-id>
     * : Suggest incoming and outgoing links for post-id
     *
     * ## EXAMPLES
     *
     *     wp backlink suggest 42
     *
     */

    public function suggest( $args, $assoc ) {
        global $backlinks_plugin;

        $posts = get_posts( $backlinks_plugin->build_post_query( [
            'post__in' => [ intval($args[0] ) ],
        ] ) );

		if ( !$posts ) {
			WP_CLI::error( 'No such eligible post' );
			return;
		}
        $post = $posts[0];

        $this->display_posts( __( 'This post', 'backlinks-taxonomy' ) , [ $post ] );

        $outlinks = $backlinks_plugin->get_outlinks( $post );
        $this->display_posts( __( 'Outgoing links', 'backlinks-taxonomy' ), $outlinks );

        $tag_suggest = $backlinks_plugin->backlink_suggestions_by_taxonomy( $post, 'post_tag', $outlinks );
        $this->display_posts( __( "Outgoing link suggestions based on tags", 'backlinks-taxonomy' ), $tag_suggest );

        $cat_suggest = $backlinks_plugin->backlink_suggestions_by_taxonomy( $post, 'category', array_merge( $outlinks, $tag_suggest ) );
        $this->display_posts( __( "Outgoing link suggestions based on category", 'backlinks-taxonomy' ), $cat_suggest );

        $backlinks = $backlinks_plugin->get_backlinks( $post );
        $this->display_posts( __( "Incoming links", 'backlinks-taxonomy' ), $backlinks );

        $taxonomies = get_taxonomies( [
            'public' => 1,
            'show_ui' => 1,
            'object_type' => ['post'],
        ] );

        $exclude = $backlinks;
        foreach ( $taxonomies as $taxonomy ) {
            $suggestions = $backlinks_plugin->backlink_suggestions_by_taxonomy( $post, $taxonomy, $exclude );
            $this->display_posts( __( "Incoming link suggestions based on $taxonomy", 'backlinks-taxonomy' ), $suggestions );

            $exclude = array_merge( $exclude, $suggestions );
        }
    }
}

WP_CLI::add_command( 'backlinks', new BacklinksCLI() );
