<?php
/**
 * Plugin Name:     Backlinks Taxonomy
 * Plugin URI:      https://rene.seindal.dk/wordpress/backlinks
 * Description:     Maintains a taxonomy of backlinks to published posts and pages
 * Author:          René Seindal
 * Author URI:      https://rene.seindal.dk/
 * Text Domain:     backlinks-taxonomy
 * Domain Path:     /languages
 * Version:         0.22
 *
 * @package         Backlinks
 */

namespace ReneSeindal;

if ( ! defined( 'ABSPATH' ) ) exit; // Exit if accessed directly

require_once( ABSPATH . 'wp-admin/includes/taxonomy.php' );

class Backlinks {
    protected $taxonomy_name = 'backlink';
    protected $taxonomy_count = 'backlink_count';
    protected $taxonomy_post_types = [ 'post', 'page' ];
    protected $taxonomy_post_statuses = [ 'publish', 'future' ];

    /*************************************************************
     *
     * Automatic setup of actions, filters and shortcodes based on
     * available class methods
     *
     *************************************************************/

	function __construct() {
        $methods = get_class_methods( $this );
        foreach ( $methods as $method ) {
            if ( preg_match( '/^do_(.*)_shortcode$/', $method, $match ) ) {
                add_shortcode( $match[1], [ $this, $method ] );
            }
            elseif ( preg_match( '/^do_(.*)_action(_.*)?$/', $method, $match ) ) {
                $action = $match[1];
                $extra = $match[2] ?? '';
                $prio = $this->callback_param( "$action$extra", 'priority', 10);
                $argc = $this->callback_param( $action, 'arguments', 1);
                add_action( $action, [ $this, $method ], $prio, $argc );
            }
            elseif ( preg_match( '/^do_(.*)_filter(_.*)?$/', $method, $match ) ) {
                $filter = $match[1];
                $extra = $match[2] ?? '';
                $prio = $this->callback_param( "$filter$extra", 'priority', 10);
                $argc = $this->callback_param( $filter, 'arguments', 1);
                add_filter( $filter, [ $this, $method ], $prio, $argc );
            }
        }

        $this->taxonomy_post_statuses = $this->filter_option_list( 'post_statuses',
                                                                   get_post_stati( [], 'names' ),
                                                                   $this->taxonomy_post_statuses
        );
	}

    protected function callback_param( $name, $param, $default ) {
        $property = "{$name}_callback_${param}";
        return
            property_exists( $this, $property )
            ? intval( $this->{$property} )
            : $default;
    }

    /************************************************************************
     *
     *	Options
     *
     ************************************************************************/

    protected function get_option( $name ) {
        $option = get_option( 'backlinks_' . $name );
        if ( !empty( $option ) ) return $option;

        /* In case we do a settings page one day

        $options = get_option( 'backlinks_options' );
        if ( empty($options) ) return NULL;
        if ( array_key_exists( $name, $options ) ) return $options[$name];
        if ( array_key_exists( "backlinks_$name", $options ) ) return $options["backlinks_$name"];
        */

        return NULL;
    }

    protected function has_option( $name ) {
        return !empty( $this->get_option( $name ) );
    }

    function filter_option_list( $option, $valid_values, $default = [] ) {
        $input = $this->get_option( $option );
        $output = [];

        if ( ! empty( $input ) ) {
            $output = array_filter( array_map( 'trim', explode( ',', $input ) ),
                                   function ( $v ) use ( $valid_values ) {
                                       return !empty( $v ) && in_array( $v, $valid_values);
                                   });
        }

        if ( empty( $output ) )
            $output = $default;

        return array_values( $output );
    }


    /************************************************************************
     *
     *	Helpers
     *
     ************************************************************************/

    public function post_type_ok( $post ) {
        $post = get_post( $post );
        return in_array( $post->post_type, $this->taxonomy_post_types);
    }

    public function term_to_post_id( $slug ) {
        return intval( substr( $slug, 1 ) );
    }

    public function post_id_to_term( $post_id ) {
        return "p$post_id";
    }

    function build_post_query( $query = [] ) {
        return array_merge(
            [
                'post_type' => $this->taxonomy_post_types,
                'post_status' => $this->taxonomy_post_statuses,
                'numberposts' => -1,
            ],
            $query
        );
    }

    function get_posts_query( $query ) {
        return get_posts( $this->build_post_query( $query ) );
    }

    function get_backlinks( $post ) {
        $post = get_post( $post );

        return $this->get_posts_query( [
            'tax_query' => [
                [
                    'taxonomy' => $this->taxonomy_name,
                    'field'    => 'slug',
                    'terms'    => $this->post_id_to_term( $post->ID ),
                ]
            ],
        ] );
    }

    function get_posts_with_no_backlinks() {
        return $this->get_posts_query( [
            'tax_query' => [
                [
                    'taxonomy' => $this->taxonomy_count,
                    'operator' => 'NOT EXISTS',
                ]
            ],
        ] );
    }

    function get_outlinks( $post ) {
        $post = get_post( $post );

        $terms = get_the_terms( $post, $this->taxonomy_name );

        if ( $terms === false || is_wp_error( $terms ) )
            return [];

        $ids = array_map( [ $this, 'term_to_post_id' ], wp_list_pluck( $terms, 'slug' ));

        return $this->get_posts_query( [ 'post__in' => $ids ] );
    }


    /************************************************************************
     *
     *	Helpers for CLI class
     *
     ************************************************************************/

    public function post_types() {
        return $this->taxonomy_post_types;
    }

    public function post_statuses() {
        return $this->taxonomy_post_statuses;
    }

    public function taxonomy() {
        return $this->taxonomy_name;
    }

    public function textdomain() {
        return 'backlinks-taxonomy';
    }

    /************************************************************************
     *
     *	The backlink taxonomies
     *
     ************************************************************************/

    public function do_init_action_setup_taxonomy() {
        $this->taxonomy_post_types = $this->filter_option_list( 'post_types',
                                                                get_post_types( [ 'public' => true ], 'names' ),
                                                                $this->taxonomy_post_types
        );


        register_taxonomy( $this->taxonomy_name,
                           $this->taxonomy_post_types,
                           [
                               'label' => __( 'Backlinks', 'backlinks-taxonomy' ),
                               'public' => FALSE,
                               'hierarchical' => FALSE,
                               'show_in_rest' => TRUE,
                               'publicly_queryable' => TRUE,
                               'meta_box_cb' => FALSE,
                           ] );

        $public = false;
        if ( defined( 'WP_CLI' ) && WP_CLI ) {
            $public = true;
        }

        register_taxonomy( $this->taxonomy_count,
                           $this->taxonomy_post_types,
                           [
                               'hierarchical'          => false,
                               'public'                => $public,
                               'show_in_nav_menus'     => false,
                               'show_ui'               => false,
                               'show_admin_column'     => false,
                               'query_var'             => true,
                               'rewrite'               => true,

                               'labels'                => [
                                   'name'                       => __( 'Backlink counts', 'backlinks-taxonomy' ),
                                   'singular_name'              => _x( 'Backlink count', 'taxonomy general name', 'backlinks-taxonomy' ),
                                   'search_items'               => __( 'Search Backlink counts', 'backlinks-taxonomy' ),
                                   'popular_items'              => __( 'Popular Backlink counts', 'backlinks-taxonomy' ),
                                   'all_items'                  => __( 'All Backlink counts', 'backlinks-taxonomy' ),
                                   'parent_item'                => __( 'Parent Backlink count', 'backlinks-taxonomy' ),
                                   'parent_item_colon'          => __( 'Parent Backlink count:', 'backlinks-taxonomy' ),
                                   'edit_item'                  => __( 'Edit Backlink count', 'backlinks-taxonomy' ),
                                   'update_item'                => __( 'Update Backlink count', 'backlinks-taxonomy' ),
                                   'view_item'                  => __( 'View Backlink count', 'backlinks-taxonomy' ),
                                   'add_new_item'               => __( 'Add New Backlink count', 'backlinks-taxonomy' ),
                                   'new_item_name'              => __( 'New Backlink count', 'backlinks-taxonomy' ),
                                   'separate_items_with_commas' => __( 'Separate Backlink counts with commas', 'backlinks-taxonomy' ),
                                   'add_or_remove_items'        => __( 'Add or remove Backlink counts', 'backlinks-taxonomy' ),
                                   'choose_from_most_used'      => __( 'Choose from the most used Backlink counts', 'backlinks-taxonomy' ),
                                   'not_found'                  => __( 'No Backlink counts found.', 'backlinks-taxonomy' ),
                                   'no_terms'                   => __( 'No Backlink counts', 'backlinks-taxonomy' ),
                                   'menu_name'                  => __( 'Backlink counts', 'backlinks-taxonomy' ),
                                   'items_list_navigation'      => __( 'Backlink counts list navigation', 'backlinks-taxonomy' ),
                                   'items_list'                 => __( 'Backlink counts list', 'backlinks-taxonomy' ),
                                   'most_used'                  => _x( 'Most Used', 'backlink count', 'backlinks-taxonomy' ),
                                   'back_to_items'              => __( '&larr; Back to Backlink counts', 'backlinks-taxonomy' ),
                               ],
                               'show_in_rest'          => false,

                               'update_count_callback' => '_update_generic_term_count',
                           ] );

    }

    /*
     * Add filter on post lists on backlink_count taxonomy
     */

    protected $restrict_manage_posts_callback_arguments = 2;

    function do_restrict_manage_posts_action_backlink_count( $post_type, $which ) {
        $tax = get_taxonomy( $this->taxonomy_count );


        if ( in_array( $post_type, $tax->object_type ) ) {
            $var = $this->taxonomy_count;

            $selected = NULL;
            if ( isset( $_GET[$var] ) && is_string( $_GET[$var] ) ) {
                $selected = sanitize_title( $_GET[$var] );
            }

            wp_dropdown_categories(array(
                'show_option_all' =>  $tax->labels->all_items,
                'taxonomy'        =>  $this->taxonomy_count,
                'value_field'     => 'slug',
                'name'            =>  $this->taxonomy_count,
                'orderby'         =>  'name',
                'selected'        =>  $selected,
                'hierarchical'    =>  false,
                'show_count'      =>  true,
                'hide_empty'      =>  false,
            ) );
        }
    }

    /************************************************************************
     *
     *	Backlinks column in admin posts list
     *
     ************************************************************************/

    // Add the column

    protected $manage_posts_columns_callback_arguments = 2;

    function do_manage_posts_columns_filter( $columns, $post_type ) {
        $tax = get_taxonomy( $this->taxonomy_name );

        if ( in_array( $post_type, $tax->object_type ) ) {
            $columns[ 'backlink_count' ] = __( 'Backlinks', 'backlinks-taxonomy' );
        }
        return $columns;
    }

    function do_manage_pages_columns_filter( $columns ) {
        return $this->do_manage_posts_columns_filter( $columns, 'page' );
    }

    // Display the column content

    protected $manage_posts_custom_column_callback_arguments = 2;

    function do_manage_posts_custom_column_action( $column, $post_id) {
        $tax = get_taxonomy( $this->taxonomy_name );

        if ( $column == 'backlink_count' ) {
            $post = get_post( $post_id );
            if ( in_array( $post->post_status, $this->taxonomy_post_statuses ) ) {
                $count = count( $this->get_backlinks( $post ) );
                if ( $count == 0 ) {
                    _e( 'Has no backlinks', 'backlinks-taxonomy' );
                } else {
                    $url = admin_url( 'edit.php?backlink_list=' . $post_id );
                    $text = sprintf( _n( 'Has %d backlink', 'Has %d backlinks', $count, 'backlinks-taxonomy' ), $count );
                    printf( '<a href="%1$s">%2$s</a>', esc_url( $url ), esc_html( $text ) );
                }

                echo "<br/>";

                $count = count( $this->get_outlinks( $post ) );
                if ( $count == 0 ) {
                    _e( 'Has no outlinks', 'backlinks-taxonomy' );
                } else {
                    $url = admin_url( 'edit.php?backlink_list=-' . $post_id );
                    $text = sprintf( _n( 'Has %d outlinks', 'Has %d outlinks', $count, 'backlinks-taxonomy' ), $count );
                    printf( '<a href="%1$s">%2$s</a>', esc_url( $url ), esc_html( $text ) );
                }

            } else {
                _e( 'n/a', 'backlinks-taxonomy' );
            }
        }
    }

    protected $manage_pages_custom_column_callback_arguments = 2;

    function do_manage_pages_custom_column_action( $column, $post_id) {
        $this->do_manage_posts_custom_column_action( $column, $post_id );
    }


    // Make column sortable

    protected $init_action_backlink_column_sortable_priority = 20;

    function do_init_action_backlink_column_sortable() {
        foreach ( $this->taxonomy_post_types as $post_type ) {
            add_filter( "manage_edit-{$post_type}_sortable_columns", function ( $columns ) {
                $columns['backlink_count'] = 'backlink_count';
                return $columns;
            } );
        }
    }

    function do_pre_get_posts_action_backlink_sorting( $query ) {
        if ( ! is_admin() ) return;

        $orderby = $query->get( 'orderby' );
        if ( $orderby == 'backlink_count' ) {
            $query->set( 'meta_key', '_backlinks_count' );
            $query->set( 'orderby', 'meta_value_num' );
        }
    }


    // Add query_var for column link

    function do_init_action_backlink_column( ) {
        global $wp;
        $wp->add_query_var('backlink_list');
    }

    // Intercept the search for the column link
    // query_arg 'backlink_list' is post ID
    // - if negative show outlinks instead of incoming links

    function do_pre_get_posts_action_backlink_list( $query ) {
        if ( ! ( is_admin() && $query->is_main_query() ) )
            return;

        $id = $query->get( 'backlink_list' );
        if ( ! $id )
            return;

        // This is a backlinks_list search
        // Show nothing if succesive tests fail
        $query->set( 'post__in', [ 0 ] );

        $id = intval( $id );
        if ( ! $id )
            return;

        $post = get_post( abs( $id ) );
        if ( ! $post )
            return;

        if ( $id > 0 ) {
            $links = $this->get_backlinks( $post );
        } else {
            $links = $this->get_outlinks( $post );
        }

        $query->set( 'post__in', wp_list_pluck( $links, 'ID' ) );
    }

    /************************************************************************
     *
     *	Backlink suggestions
     *
     ************************************************************************/

    public function backlink_suggestions_by_taxonomy( $post, $taxonomy, $exclude = [] ) {
        $terms = get_the_terms( $post, $taxonomy );

        if ( $terms === false || is_wp_error( $terms ) )
            return [];

        $count = wp_count_posts();
        $limit = intval( $count->publish  / 3 );

        $terms = array_values( array_filter( $terms, function ( $t ) use ($limit) { return ($t->count < $limit); } ) );

        $posts = get_posts( $this->build_post_query( [
            'tax_query' => [
                [
                    'taxonomy' => $taxonomy,
                    'field'    => 'id',
                    'terms'    => wp_list_pluck( $terms, 'term_id' ),
                ]
            ],
            'post__not_in' => array_merge( [ $post->ID ], wp_list_pluck( $exclude, 'ID' ) ),
        ] ) );

        return $posts;
    }

    function do_init_action_backlink_suggestions( ) {
        global $wp;
        $wp->add_query_var('backlink_suggest');
    }

    function do_pre_get_posts_action_backlink_suggestions( $query ) {
        if ( ! ( is_admin() && $query->is_main_query() ) )
            return;

        $id = $query->get( 'backlink_suggest' );
        if ( ! $id )
            return;

        // This is a backlinks_suggest search
        // Show nothing if succesive tests fail
        $query->set( 'post__in', [ 0 ] );

        $id = intval( $id );
        if ( ! $id )
            return;

        $post = get_post( $id );
        if ( ! $post )
            return;

        $backlinks = $this->get_backlinks( $post );

        $taxonomies = get_taxonomies( [
            'public' => 1,
            'show_ui' => 1,
            'object_type' => [ $post->post_type ],
        ] );

        $suggestions = [];
        foreach ( $taxonomies as $taxonomy ) {
            $taxonomy_suggest = $this->backlink_suggestions_by_taxonomy( $post, $taxonomy, array_merge( $backlinks, $suggestions ) );
            $suggestions = array_merge( $suggestions, $taxonomy_suggest );
        }

        $query->set( 'post__in', wp_list_pluck( $suggestions, 'ID' ) );
    }

    protected $post_row_actions_callback_arguments = 2;

    function do_post_row_actions_filter_backlink_suggestions( $actions, $post ) {
        if ( $this->post_type_ok( $post ) ) {
            $url = admin_url( 'edit.php?backlink_suggest=' . $post->ID );

            $actions[ 'backlink_suggest' ] = sprintf( '<a href="%1$s">%2$s</a>',
                                                      esc_url( $url ),
                                                      esc_html( __( 'Suggest backlinks', 'backlinks-taxonomy' ) )
            );
        }

        return $actions;
    }

    /************************************************************************
     *
     *	Content parser
     *
     ************************************************************************/

    protected function content_link_parser( $post ) {
        $post = get_post( $post );

        $links = [];

        $input = mb_convert_encoding( $post->post_content, 'HTML-ENTITIES', 'UTF-8' );

        $xml = new \DOMDocument( '1.0', 'UTF-8' );
        $error_mode = libxml_use_internal_errors( TRUE );

        if ( $xml->LoadHTML( "<html>$input</html>", LIBXML_HTML_NOIMPLIED | LIBXML_HTML_NODEFDTD ) ) {
            $anchors = $xml->getElementsByTagName( 'a' );

            foreach ( $anchors as $anchor ) {
                $href = $anchor->getAttribute( 'href' );
                if ( $href ) {
                    $id = url_to_postid( $href );

                    if ( $id ) {
                        $links[$id] = $href;
                    }
                }

            }
        }

        libxml_use_internal_errors($error_mode);

        return array_keys( $links );
    }

    /************************************************************************
     *
     *	Register and deregister outgoing links for post/page
     *
     ************************************************************************/

    // $targets - array of post_ids
    public function update_taxonomy( $post, $targets ) {
        $post = get_post( $post );

        wp_set_post_terms( $post->ID,
                           array_map( [ $this, 'post_id_to_term' ], $targets ),
                           $this->taxonomy_name,
                           false
        );
    }

    private function backlink_count_name( $count = 0 ) {
        if ( 0 == $count )
            $slug = __( 'Has 0 backlinks', 'backlinks-taxonomy' );
        elseif ( $count < 10 )
            $slug = sprintf( _n( 'Has %d backlink', 'Has %d backlinks', $count, 'backlinks-taxonomy' ), $count );
        else
            $slug = __( 'Has many backlinks', 'backlinks-taxonomy' );

        return $slug;
    }

    public function update_taxonomy_count( $post ) {
        $post = get_post( $post );

        $inbound = $this->get_backlinks( $post );
        $count = count( $inbound );

        $slug = $this->backlink_count_name( $count );
        wp_set_post_terms( $post->ID, $slug, $this->taxonomy_count, false );

        // This is needed for sorting post lists on backlink count
        update_post_meta( $post->ID, '_backlinks_count', $count ?? 'none' );
    }

    public function register_outlinks( $post ) {
        $post = get_post( $post );
        $outlinks = $this->content_link_parser( $post );

        $this->update_taxonomy( $post, $outlinks );
        $this->update_taxonomy_count( $post );

        foreach ( $outlinks as $link )
            $this->update_taxonomy_count( $link );

        $unlinked = $this->get_posts_with_no_backlinks();
        foreach ( $unlinked as $link )
            $this->update_taxonomy_count( $link );

        return $outlinks;
    }

    public function deregister_outlinks( $post ) {
        $post = get_post( $post );
        $this->update_taxonomy( $post->ID, [] );
    }


    protected $transition_post_status_callback_arguments = 3;

    function do_transition_post_status_action_register_outlinks( $new, $old, $post ) {
        // Don't bother parsing unpublished posts
        if ( !in_array( $new, $this->taxonomy_post_statuses ) )
            return;

        if ( ! $this->post_type_ok( $post ) ) {
            return;
        }

        $transient = $this->taxonomy_name . '_' . $post->ID;
        $lasttime = get_transient( $transient );
        if ( $lasttime && $lasttime == $post->post_modified )
            return;
        set_transient( $transient, $post->post_modified, WEEK_IN_SECONDS );

        $this->register_outlinks( $post );
    }
}

global $backlinks_plugin;
$backlinks_plugin = new Backlinks();

if ( defined( 'WP_CLI' ) && WP_CLI ) {
    require_once dirname( __FILE__ ) . '/backlinks-cli.php';
}
